"""
    File name: field.py
    Author: joeschmid
    Date created: 4/9/17
"""
import json

from .base_generator import BaseGenerator

DEFAULT_TYPE = 'string'


class FieldType(object):
    """Enum-style class used to specify known Field types"""
    DIMENSION, DIMENSION_GROUP, FILTER, MEASURE = range(1, 5)

    @classmethod
    def type_name(cls, type_id):
        if type_id == cls.DIMENSION:
            return 'dimension'
        elif type_id == cls.DIMENSION_GROUP:
            return 'dimension_group'
        elif type_id == cls.FILTER:
            return 'filter'
        elif type_id == cls.MEASURE:
            return 'measure'
        else:
            raise ValueError('Type {} is not a valid FieldType'.
                             format(type_id))


class Field(BaseGenerator):
    """Base class used to generate fields within a
    :class:`~lookmlgen.view.View`

    :param field_type: Name of the view
    :param name: Name of the field
    :param type: Type of the field contents, e.g. string, number, etc.
    :param label: Label to use when displaying the field
    :param sql: SQL snippet for the field
    :param hidden: Flag to designate the field as hidden
    :param file: File handle of a file open for writing or a StringIO object
    :param group_label: Group label to use for grouping the field
    :param description: Field description that is show if a user hovers over the help link in the field picker
    :type field_type: a class variable from :class:`FieldType`
    :type name: string
    :type type: string
    :type label: string
    :type sql: string
    :type hidden: bool
    :type file: File handle or StringIO object
    :type group_label: string
    :type description: string

    """
    def __init__(self, field_type, name, type=DEFAULT_TYPE, label=None,
                 sql=None, hidden=None, file=None, group_label=None, description=None, **kwargs):
        super(Field, self).__init__(file=file)
        self.field_type = field_type
        self.type_name = FieldType.type_name(field_type)
        self.name = name
        self.type = type
        self.label = label
        self.group_label = group_label
        self.sql = sql if sql else '${TABLE}.%s' % name
        self.hidden = hidden
        self.description = description

    def generate_lookml(self, file=None, format_options=None):
        """ Writes LookML for a field to a file or StringIO buffer.

        :param file: File handle of a file open for writing or a
                     StringIO object
        :param format_options: Formatting options to use during generation
        :type file: File handle or StringIO object
        :type format_options:
            :class:`~lookmlgen.base_generator.GeneratorFormatOptions`

        """
        f = file if file else self.file
        fo = format_options if format_options else self.format_options
        f.write('{indent}{self.type_name}: {self.name} {{\n'.
                format(indent=' ' * fo.indent_spaces, self=self))
        if self.hidden:
            f.write('{indent}hidden: yes\n'.
                    format(indent=' ' * 2 * fo.indent_spaces))
        if self.label:
            f.write('{indent}label: "{self.label}"\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))
        if self.group_label:
            f.write('{indent}group_label: "{self.group_label}"\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))

        if self.description:
            f.write('{indent}description: "{self.description}"\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))

        if self.type and not (fo.omit_default_field_type and
                              self.type == DEFAULT_TYPE):
            f.write('{indent}type: {self.type}\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))
        self._generate(f, fo)
        if self.sql:
            f.write('{indent}sql: {self.sql} ;;\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))
        f.write('{indent}}}\n'.format(indent=' ' * fo.indent_spaces))
        return

    def _generate(self, f, fo):
        return


class Dimension(Field):
    """Generates LookML for a dimension field in a
    :class:`~lookmlgen.view.View`

    :param name: Name of the dimension
    :param primary_key: Flag to designate the field as a primary key
    :type name: string
    :type primary_key: bool

    """
    def __init__(self, name, primary_key=None, **kwargs):
        super(Dimension, self).__init__(FieldType.DIMENSION, name, **kwargs)
        self.primary_key = primary_key

    def _generate(self, f, fo):
        if self.primary_key:
            f.write('{indent}primary_key: yes\n'.
                    format(indent=' ' * 2 * fo.indent_spaces))


class DimensionGroup(Field):
    """Generates LookML for a dimension_group field in a
        :class:`~lookmlgen.view.View`

    :param name: Name of the dimension group
    :param timeframes: Timeframes for the group
    :param datatype: Datatype for the group, defaults to 'datetime'
    :type name: string
    :type timeframes: list of strings
    :type datatype: string

    """
    def __init__(self, name, timeframes=None, datatype='datetime', **kwargs):
        super(DimensionGroup, self).__init__(FieldType.DIMENSION_GROUP, name,
                                             type='time', **kwargs)
        self.timeframes = timeframes
        self.datatype = datatype

    def _generate(self, f, fo):
        if not self.timeframes and not fo.omit_time_frames_if_not_set:
            self.timeframes = ['time', 'date', 'week', 'month']

        if self.timeframes:
            f.write('{indent}timeframes: {timeframes}\n'.
                    format(indent=' ' * 2 * fo.indent_spaces,
                           timeframes=json.dumps(self.timeframes).replace('"', '')))
        if self.datatype:
            f.write('{indent}datatype: {self.datatype}\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))


class Measure(Field):
    """Generates LookML for a measure field in a
            :class:`~lookmlgen.view.View`

    :param name: Name of the measure
    :type name: string

    """
    def __init__(self, name, **kwargs):
        super(Measure, self).__init__(FieldType.MEASURE, name, **kwargs)


class Filter(Field):
    """Generates LookML for a filter field in a
            :class:`~lookmlgen.view.View`

    :param name: Name of the filter
    :type name: string

    """
    def __init__(self, name, **kwargs):
        super(Filter, self).__init__(FieldType.FILTER, name, **kwargs)



"""
    File name: view.py
    Author: joeschmid
    Date created: 4/8/17
"""
import json
from collections import OrderedDict
try:
    from textwrap import indent
except ImportError:
    from .util import indent

from .base_generator import BaseGenerator
from .field import FieldType


class View(BaseGenerator):
    """Generates a LookML View

    Initialize a View object with your parameters,
    add Fields such as :class:`~lookmlgen.field.Dimension`,
    :class:`~lookmlgen.field.Measure`,
    :class:`~lookmlgen.field.DimensionGroup`, and
    :class:`~lookmlgen.field.Filter`, and then
    generate LookML for the view using :py:meth:`~View.generate_lookml`

    :param name: Name of the view
    :param label: Label to use for the view (may contain spaces)
    :param sql_table_name: Name of the SQL table to use in the view
    :param file: File handle of a file open for writing or a
                 StringIO object
    :type name: string
    :type label: string
    :type sql_table_name: list of strings
    :type file: File handle or StringIO object

    """
    def __init__(self, name, label=None, sql_table_name=None, file=None):
        super(View, self).__init__(file=file)
        self.name = name
        self.label = label
        self.sql_table_name = sql_table_name
        self.fields = OrderedDict()
        self.derived_table = None

    def generate_lookml(self, file=None, format_options=None):
        """ Writes LookML for the view to a file or StringIO buffer.

        :param file: File handle of a file open for writing or a
                     StringIO object
        :param format_options: Formatting options to use during generation
        :type file: File handle or StringIO object
        :type format_options:
            :class:`~lookmlgen.base_generator.GeneratorFormatOptions`

        """
        if not file and not self.file:
            raise ValueError('Must provide a file in either the constructor '
                             'or as a parameter to generate_lookml()')
        f = file if file else self.file
        fo = format_options if format_options else self.format_options
        if fo.warning_header_comment:
            f.write(fo.warning_header_comment)
        f.write('view: {self.name} {{\n'.format(self=self))
        if self.sql_table_name:
            f.write('{indent}sql_table_name: {self.sql_table_name} ;;\n'.
                    format(indent=' ' * fo.indent_spaces, self=self))
        if self.label:
            f.write('{indent}label: "{self.label}"\n'.
                    format(indent=' ' * fo.indent_spaces, self=self))

        if fo.newline_between_items:
            f.write('\n')

        if self.derived_table:
            self.derived_table.generate_lookml(file=f, format_options=fo)
            if fo.newline_between_items:
                f.write('\n')

        if fo.view_fields_alphabetical:
            ordered_fields = sorted(self.fields.items())
        else:
            ordered_fields = self.fields.items()
        self._gen_fields(f, fo, ordered_fields, [FieldType.FILTER])
        self._gen_fields(f, fo, ordered_fields,
                         [FieldType.DIMENSION, FieldType.DIMENSION_GROUP])
        self._gen_fields(f, fo, ordered_fields, [FieldType.MEASURE])

        f.write('}\n')
        return

    def add_field(self, field):
        """Adds a :class:`~lookmlgen.field.Field` object to a :class:`View`"""
        self.fields[field.name] = field
        return

    def set_derived_table(self, derived_table):
        """Adds a :class:`~lookmlgen.view.DerivedTable` object to a
         :class:`View`
        """
        self.derived_table = derived_table

    @classmethod
    def _gen_fields(cls, f, fo, sorted_fields, field_types):
        first = True
        for k, d in sorted_fields:
            if not first and fo.newline_between_items:
                f.write('\n')
            if d.field_type not in field_types:
                continue
            d.generate_lookml(file=f, format_options=fo)
            if first:
                first = False


class DerivedTable(BaseGenerator):
    """Generates the LookML View parameters to support derived
    tables, including persistent derived tables (PDTs).

    :param sql: SQL statement to execute
    :param sql_trigger_value: SQL to determine when to trigger build
    :param indexes: List of coluxn names to use as indexes
    :param file: File handle of a file open for writing or a StringIO object
    :type sql: string
    :type sql_trigger_value: string
    :type indexes: list of strings
    :type file: File handle or StringIO object

    """
    def __init__(self, sql, sql_trigger_value=None, indexes=None, file=None):
        super(DerivedTable, self).__init__(file=file)
        self.sql = sql
        self.sql_trigger_value = sql_trigger_value
        self.indexes = indexes

    def generate_lookml(self, file=None, format_options=None):
        """ Writes LookML for a derived table to a file or StringIO buffer.

        :param file: File handle of a file open for writing or a
                     StringIO object
        :param format_options: Formatting options to use during generation
        :type file: File handle or StringIO object
        :type format_options:
            :class:`~lookmlgen.base_generator.GeneratorFormatOptions`

        """
        if not file and not self.file:
            raise ValueError('Must provide a file in either the constructor '
                             'or as a parameter to generate_lookml()')
        f = file if file else self.file
        fo = format_options if format_options else self.format_options
        f.write('{indent}derived_table: {{\n'.
                format(indent=' ' * fo.indent_spaces))
        if self.sql:
            final_sql = ' ' + self.sql if '\n' not in self.sql \
                else '\n' + indent(self.sql, ' ' * 3 * fo.indent_spaces)
            f.write('{indent}sql:{sql} ;;\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, sql=final_sql))
        if self.sql_trigger_value:
            f.write('{indent}sql_trigger_value: '
                    '{self.sql_trigger_value} ;;\n'.
                    format(indent=' ' * 2 * fo.indent_spaces, self=self))
        if self.indexes:
            f.write('{indent}indexes: {indexes}\n'.
                    format(indent=' ' * 2 * fo.indent_spaces,
                           indexes=json.dumps(self.indexes)))
        f.write('{indent}}}\n'.format(indent=' ' * fo.indent_spaces))



"""
    File name: test_view.py
    Author: joeschmid
    Date created: 4/17/17
"""
import os
import six

from lookmlgen import view
from lookmlgen import field
from lookmlgen import base_generator


test_format_options = base_generator.\
    GeneratorFormatOptions(warning_header_comment=None)


def test_basic_view():
    testname = 'basic_view'
    v = view.View(testname)
    v.add_field(field.Dimension('dimension1', sql='${TABLE}.dim1'))
    f = six.StringIO()
    v.generate_lookml(f, format_options=test_format_options)
    lookml = f.getvalue()
    with open(os.path.join(os.path.dirname(__file__),
                           'expected_output/%s.lkml' % testname),
              'rt') as expected:
        assert lookml == expected.read()


def test_pdt_view():
    testname = 'pdt_view'
    pdt = view.DerivedTable(sql="SELECT id, count(*) c FROM table GROUP BY id",
                            sql_trigger_value='DATE()',
                            indexes=['id'])
    v = view.View(testname)
    v.derived_table = pdt
    v.add_field(field.Dimension('id', type='number',
                                primary_key=True))
    v.add_field(field.Dimension('c', type='number'))
    v.add_field(field.Measure('sum_c', sql='${TABLE}.c', type='sum'))
    f = six.StringIO()
    v.generate_lookml(f, format_options=test_format_options)
    lookml = f.getvalue()
    six.print_(lookml)
    with open(os.path.join(os.path.dirname(__file__),
                           'expected_output/%s.lkml' % testname),
              'rt') as expected:
        assert lookml == expected.read()


def test_dimension_group():
    testname = 'dimension_group_test'
    v = view.View(testname)
    v.add_field(field.DimensionGroup('dimension1', sql='${TABLE}.dim1'))
    f = six.StringIO()
    v.generate_lookml(f, format_options=test_format_options)
    lookml = f.getvalue()
    six.print_(lookml)
    with open(os.path.join(os.path.dirname(__file__),
                           'expected_output/%s.lkml' % testname),
              'rt') as expected:
        assert lookml == expected.read()


def test_dimension_group_no_timeframes():
    testname = 'dimension_group_no_timeframes_test'
    v = view.View(testname)
    v.add_field(field.DimensionGroup('dimension1', sql='${TABLE}.dim1'))
    f = six.StringIO()
    fo_omit_timeframes = base_generator.\
        GeneratorFormatOptions(warning_header_comment=None, omit_time_frames_if_not_set=True)
    v.generate_lookml(f, format_options=fo_omit_timeframes)
    lookml = f.getvalue()
    six.print_(lookml)
    with open(os.path.join(os.path.dirname(__file__),
                           'expected_output/%s.lkml' % testname),
              'rt') as expected:
        assert lookml == expected.read()



"""
    File name: util.py
    Author: joeschmid
    Date created: 4/16/17
"""


def indent(s, num_spaces):
    lines = s.splitlines()
    indented = [' ' * num_spaces + l for l in lines]
    # indented[0] = lines[0]
    if s[-1] == '\n':
        indented[-1] += '\n'
    indented = '\n'.join(indented)
    return indented



"""
    File name: base_generator.py
    Author: joeschmid
    Date created: 4/8/17
"""
import abc
import six

DEFAULT_WARNING_HEADER_COMMENT = \
    """# STOP! This file was generated by an automated process.
# Any edits you make will be lost the next time it is
# re-generated.\n"""


class GeneratorFormatOptions(object):
    """Specify formatting options to be used during LookML generation

    :param indent_spaces: Number of spaces to indent
    :param newline_between_items: Add a newline between items
    :param omit_default_field_type: Leave out 'type: string'
    :param warning_header_comment: Text to use as a comment as the top
                                   of the file warning the user that the
                                   file will get overwritten
    :param omit_time_frames_if_not_set: If no time frame is specified for
                                        a dimension_group field, omit the
                                        time_frames parameter. If timeframes
                                        is not included every timeframe
                                        option will be added to the dimension group.
    :type indent_spaces: int
    :type newline_between_items: bool
    :type omit_default_field_type: bool
    :type warning_header_comment: string
    :type omit_time_frames_if_not_set: bool

    """
    def __init__(self, indent_spaces=2, newline_between_items=True,
                 omit_default_field_type=True, view_fields_alphabetical=True,
                 warning_header_comment=DEFAULT_WARNING_HEADER_COMMENT,
                 omit_time_frames_if_not_set=False):
        self.indent_spaces = indent_spaces
        self.newline_between_items = newline_between_items
        self.omit_default_field_type = omit_default_field_type
        self.warning_header_comment = warning_header_comment
        self.view_fields_alphabetical = view_fields_alphabetical
        self.omit_time_frames_if_not_set = omit_time_frames_if_not_set


@six.add_metaclass(abc.ABCMeta)
class BaseGenerator:
    """ Abstract base class for any subclass that generates LookML

    :param file: File handle of a file open for writing or a
                 StringIO object
    :param format_options: Formatting options to use during generation
    :type file: File handle or StringIO object
    :type format_options:
        :class:`~lookmlgen.base_generator.GeneratorFormatOptions`

    """
    def __init__(self, file=None, format_options=GeneratorFormatOptions()):
        self.file = file
        self.format_options = format_options

    @abc.abstractmethod
    def generate_lookml(self, file=None, format_options=None):
        """ Implement this method in subclasses to generate LookML

        :param file: File handle of a file open for writing or a
                     StringIO object
        :param format_options: Formatting options to use during generation
        :type file: File handle or StringIO object
        :type format_options:
            :class:`~lookmlgen.base_generator.GeneratorFormatOptions`

        """
        raise NotImplementedError(
            'You must implement the generate_lookml() method')

    @classmethod
    def __subclasshook__(cls, C):
        if cls is BaseGenerator:
            if any("generate_lookml" in B.__dict__ for B in C.__mro__):
                return True
        return NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'six>=1.10.0'
]

test_requirements = [
    'pip==9.0.1'
    'bumpversion==0.5.3'
    'backports.tempfile==1.0rc1'
    'wheel==0.29.0'
    'watchdog==0.8.3'
    'flake8==3.3.0'
    'tox==2.7.0'
    'coverage==4.3.4'
    'six==1.10.0'
    'Sphinx==1.5.5'
    'cryptography==1.8.1'
    'PyYAML==3.12'
    'pytest==3.0.7'
]

setup(
    name='lookml-gen',
    version='0.1.8',
    description="Programmatically generate LookML",
    long_description=readme + '\n\n' + history,
    author="Joe Schmid",
    author_email='jschmid@symphonyrm.com',
    url='https://github.com/symphonyrm/lookml-gen',
    packages=[
        'lookmlgen',
    ],
    package_dir={'lookmlgen':
                 'lookmlgen'},
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='lookml-gen',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
